__all__ = (
    'MroSlotsBase',
    'MroDictBase',
    'MroSlotsSub',
    'mro_slots_sub',

    'MroAsync',
    'mro_async',

    'MroData',
    'MroDataDictMix',
    'MroDataDictSlotMix',
    'mro_test_data_dict_slot_mix',

    'MroValuesBase',
    'MroValuesSub',
    'mro_values_sub',

    'DataAnnotations',
)
from dataclasses import dataclass
from dataclasses import field
from dataclasses import InitVar
from functools import cached_property
from types import GenericAlias
from typing import Any
from typing import ClassVar
from typing import Final
from typing import Literal
from typing import Optional
from typing import Union


class MroSlotsBase:
    __slots__ = ('mro_slots_base', )

    def __init__(self):
        self.mro_slots_base = 'mro_slots_base'


class MroDictBase(MroSlotsBase):

    def __init__(self):
        super().__init__()
        self.mro_dict_base = 'mro_dict_base'


class MroSlotsSub(MroDictBase):
    __slots__ = ('mro_slots_sub', )

    def __init__(self):
        super().__init__()
        self.mro_slots_sub = 'mro_slots_sub'


mro_slots_sub = MroSlotsSub()


class MroAsync:
    _async_classmethod = '_async_classmethod'
    _classmethod = '_classmethod'
    _async_method = '_async_method'
    _method = '_method'
    _cprop = '_cprop'
    _async_prop = '_async_prop'
    _prop = '_prop'
    _async_staticmethod = '_async_staticmethod'
    _staticmethod = '_staticmethod'

    @classmethod
    async def async_classmethod(cls): return cls._async_classmethod

    @classmethod
    def classmethod(cls): return cls._classmethod

    async def async_method(self): return self._async_method

    def method(self): return self._method

    @cached_property
    def cprop(self): return self._cprop

    @property
    async def async_prop(self): return self._async_prop

    @property
    def prop(self): return self._prop

    @staticmethod
    async def async_staticmethod(): return MroAsync._async_staticmethod

    @staticmethod
    def staticmethod(): return MroAsync._staticmethod


mro_async = MroAsync()


@dataclass
class MroData:
    __data = '__data'
    __dataclass_classvar__: ClassVar[str] = '__dataclass_classvar__'
    __dataclass_classvar: ClassVar[str] = '__dataclass_classvar'
    __dataclass_default_factory: Union[dict, str] = field(default_factory=dict, init=False)
    __dataclass_default_factory_init: Union[dict, str] = field(default_factory=dict)
    dataclass_classvar: ClassVar[str] = 'dataclass_classvar'
    dataclass_default_factory: Union[dict, str] = field(default_factory=dict, init=False)
    dataclass_default_factory_init: Union[dict, str] = field(default_factory=dict)
    dataclass_default: str = field(default='dataclass_default', init=False)
    dataclass_default_init: str = field(default='dataclass_default_init')
    dataclass_initvar: InitVar[str] = 'dataclass_initvar'
    dataclass_str: str = 'dataclass_integer'

    def __post_init__(self, dataclass_initvar): pass

    __class_getitem__ = classmethod(GenericAlias)


class MroDataDictMix(MroData):
    subclass_annotated_str: str = 'subclass_annotated_str'
    subclass_classvar: ClassVar[str] = 'subclass_classvar'
    subclass_str = 'subclass_str'

    def __init__(self, dataclass_initvar='dataclass_initvar_1', subclass_dynamic='subclass_dynamic'):
        super().__init__()
        super().__post_init__(dataclass_initvar=dataclass_initvar)
        self.subclass_dynamic = subclass_dynamic


class MroDataDictSlotMix(MroDataDictMix):
    __slots__ = ('_slot_property', 'slot',)

    # Add init=True dataclass attrs if it subclassed and not @dataclass
    def __init__(self, dataclass_initvar='dataclass_initvar_2', slot_property_initvar='slot_property_initvar',
                 slot_initvar='slot_initvar'):
        super().__init__()
        super().__post_init__(dataclass_initvar=dataclass_initvar)
        self._slot_property = slot_property_initvar
        self.slot = slot_initvar

    @property
    def slot_property(self):
        return self._slot_property


mro_test_data_dict_slot_mix = MroDataDictSlotMix()


class MroValuesBase:
    __slots__ = ('_hash', '_repr', '_test', 'missing')
    __ignore_hash__ = ('_repr', )
    __ignore_init__ = ('_repr', )
    __ignore_repr__ = ('_hash', 'test', )
    __ignore_state__ = ('_hash',)

    @property
    def test(self):
        return self._test


class MroValuesSub(MroValuesBase):
    __slots__ = ('_prop', '_slot', 'attr', 'init', 'init_ignore')
    __ignore_hash__ = ('_prop',)
    __ignore_init__ = ('init_ignore', )
    __ignore_repr__ = ('attr',)
    __ignore_state__ = ('attr',)

    def __init__(self, init, init_ignore='init_ignore'):
        self.init = init
        self.init_ignore = init_ignore


mro_values_sub = MroValuesSub(1)


@dataclass
class DataAnnotations:
    any: Any = 'any'
    classvar: ClassVar[str] = 'classvar'
    classvar_optional: ClassVar[Optional[str]] = 'classvar_optional'
    classvar_optional_union: ClassVar[Optional[Union[str, int]]] = 'classvar_optional_union'
    classvar_union: ClassVar[Union[str, int]] = 'classvar_union'
    final: Final = 'final'
    final_str: Final[str] = 'final_str'
    integer: int = 1
    initvar: InitVar[str] = 'initvar'
    initvar_optional: InitVar[Optional[str]] = 'initvar_optional'
    literal: Literal['literal', 'literal2'] = 'literal2'
    literal_optional: Optional[Literal['literal_optional', 'literal_optional2']] = 'literal_optional2'
    optional: Optional[str] = 'optional'
    union: Union[str, int] = 1
    optional_union: Optional[Union[str, bool]] = True

    # noinspection PyUnusedLocal
    def __post_init__(self, initvar: int, initvar_optional: Optional[int]):
        self.a = initvar



from bapy import icc

import bapy.data.frame
from bapy.data.frame import *
from bapy import *


def test_data_c():
    assert Stack(init=True) != A.init
    assert (A.init[0].file.exists and A.init[0].file.include) is True
    assert A.init[0].external and A.init[1].external is False
    assert (A.init[1].file.exists and A.init[1].file.include) is False
    assert (Stack(init=True)[1].file.exists and Stack(init=True)[1].file.include) is False

    assert len(A.c_1) > 2
    assert A.c_1[0].file.path == Path(bapy.data.frame.__file__)
    assert A.c_1[0].file.exists is True
    assert A.c_1[0].include is True
    assert A.c_1[0].cls == A.__name__
    assert classmethod.__name__ in A.c_1[0].decorator
    assert A.c_1[0].module is False
    assert A.c_1[0].qualname == A.c.__qualname__
    assert A.c_1[0].lines[A.c_1[0].lineno]['coro']['func'] is False
    assert A.c_1[0].args == dict()

    assert A.c_1[1].file.path == Path(bapy.data.frame.__file__)
    assert A.c_1[1].file.exists is True
    assert A.c_1[1].file.include is True
    assert A.c_1[1].function == N._MODULE
    assert A.c_1[1].args == dict()


def test_context():
    assert context.no_sync.caller == context.no_sync.stack[FRAME_INDEX]
    assert context.no_sync.stack[0].file.exists is True
    assert context.no_sync.stack[0].include is True
    lineno = context.no_sync.stack[0].lineno
    assert context.no_sync.stack[0].lines[lineno]['coro']['func'] is True
    assert context.no_sync.stack[0].function.decorators == list()
    assert context.no_sync.stack[0].function.name == asynccontext_call_async.__name__
    assert context.no_sync.stack[0].coro is True
    assert context.no_sync.stack[0].real == 0
    assert context.no_sync.stack[0].lines[lineno]['coro']['line'] is False
    assert context.no_sync.stack[FRAME_INDEX].file.exists is True
    assert context.no_sync.stack[FRAME_INDEX].include is True
    lineno = context.no_sync.stack[FRAME_INDEX].lineno
    assert context.no_sync.stack[FRAME_INDEX].lines[lineno]['coro']['func'] is True
    assert context.no_sync.stack[FRAME_INDEX].decorators == [asynccontextmanager.__name__]
    assert context.no_sync.stack[FRAME_INDEX].coro is True
    assert context.no_sync.stack[FRAME_INDEX].real == 0
    assert context.no_sync.stack[FRAME_INDEX].lines[lineno]['coro']['line'] is False


def test_thread():
    # icc(thread)
    assert thread[0].decorators == [staticmethod.__qualname__]
    assert thread[0].qualname == A.s.__qualname__
    assert thread[0].coro is False
    assert thread[FRAME_INDEX].real is None
    caller = thread()
    real = thread(real=True)
    assert caller == real.caller == real.real


# noinspection PyUnusedLocal
def test_completed():
    lineno = completed[0].lineno
    assert completed[0].lines[lineno]['coro']['func'] is True
    assert completed[0].decorators == [staticmethod.__qualname__]
    assert completed[0].qualname == A.d.__qualname__
    assert completed[0].coro is True
    assert completed[0].lines[lineno]['coro']['line'] is False
    icc(completed)
    caller = completed()  # 1
    assert caller.coro is True
    assert caller.path.has(asyncio.__name__)

    real = completed(real=True)
    index = completed.index(real)  # 6
    assert real.real.module is True
    assert completed[FRAME_INDEX].real is None
    caller = completed()
    real = completed(real=True)
    # icc(caller, real)
    # real.real.module is True
    # assert caller == real.caller == real.real


def test_foryield():
    pass
    # icc(foryield)
    # assert thread[0].function.qual == A.s.__qualname__
    # assert thread[0].info.ASYNC is False
    # assert thread[FRAME_INDEX].info.real is None
    # caller = thread()
    # real = thread(real=True)
    # icc(caller, real)
    # assert caller == real.caller == real.real


def test_prop_to_thread():
    pass
    # icc(prop_to_thread)
    # assert thread[0].function.qual == A.s.__qualname__
    # assert thread[0].info.ASYNC is False
    # assert thread[FRAME_INDEX].info.real is None
    # caller = thread()
    # real = thread(real=True)
    # icc(caller, real)
    # assert caller == real.caller == real.real


def test_task():
    pass
    # icc(task)
    # assert thread[0].function.qual == A.s.__qualname__
    # assert thread[0].info.ASYNC is False
    # assert thread[FRAME_INDEX].info.real is None
    # caller = thread()
    # real = thread(real=True)
    # icc(caller, real)
    # assert caller == real.caller == real.real


def test_ensure():
    pass
    # icc(ensure)
    # assert thread[0].function.qual == A.s.__qualname__
    # assert thread[0].info.ASYNC is False
    # assert thread[FRAME_INDEX].info.real is None
    # caller = thread()
    # real = thread(real=True)
    # icc(caller, real)
    # assert caller == real.caller == real.real


def test_gather():
    # gather = asyncio.run(asyncdef(gth=True), debug=False)

    pass
    # icc(gather)
    # assert thread[0].function.qual == A.s.__qualname__
    # assert thread[0].info.ASYNC is False
    # assert thread[FRAME_INDEX].info.real is None
    # caller = thread()
    # real = thread(real=True)
    # icc(caller, real)
    # assert caller == real.caller == real.real


test_data_c()
test_context()
test_thread()
test_completed()
test_foryield()
test_prop_to_thread()
test_task()
test_ensure()
test_gather()



import asyncio
from asyncio import as_completed
from asyncio import create_task
from asyncio import to_thread
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import AsyncIterable
from typing import Optional
from typing import Union

from async_property import async_property

from bapy import Executor
from bapy import Stack, Frame


@dataclass
class CallerStack:
    caller: Frame
    stack: Stack


@dataclass
class NoSyncSync:
    no_sync: CallerStack
    sync: CallerStack


class A:
    init: Stack = Stack(init=True)
    c_1: Optional[Stack] = None

    @classmethod
    def c(cls):
        cls.c_1 = Stack()
        cls.s()
        pass

    # noinspection PyMethodMayBeStatic
    def m(self) -> Stack:
        return Stack()

    @property
    def p(self):
        return 'p'

    @staticmethod
    def s() -> Stack:
        rv = Stack()
        return rv

    @async_property
    async def a(self) -> Stack:
        b = await to_thread(self.m)
        return b

    # noinspection PyPropertyDefinition
    @property
    async def n(self):
        pass

    @staticmethod
    def z():
        pass

    @staticmethod
    async def d() -> Stack:
        s = Stack()
        return s

    @classmethod
    async def e(cls):
        pass

    async def f(self):
        pass


def func():
    pass


async def asynccontext_call_async() -> CallerStack:
    s = Stack()
    rv = CallerStack(caller=s(), stack=s)
    return rv


def asynccontext_call_sync() -> CallerStack:
    s = Stack()
    rv = CallerStack(caller=s(), stack=s)
    return rv


@asynccontextmanager
async def asynccontext() -> NoSyncSync:
    no_sync = await asynccontext_call_async()
    sync = await Executor.NONE.run(asynccontext_call_sync)
    rv = NoSyncSync(no_sync=no_sync, sync=sync)
    try:
        yield rv
    finally:
        pass


async def rv_async() -> Stack:
    return Stack()


async def asyncyield() -> AsyncIterable:
    a = 1
    while a:
        a -= 1
        yield rv_async


async def asyncdef(ctx: bool = False, thrd: bool = False,
                   cmp: bool = False, fyld: bool = False, pth: bool = False, tsk: bool = False,
                   ens: bool = False, gth: bool = False) -> Union[NoSyncSync, Stack]:
    if ctx:
        async with \
                asynccontext() as rv:
            return rv

    if thrd:
        return \
            await to_thread(A.s)

    if cmp:
        result = list()
        for coro in as_completed([A.d(),
                                  A.d()]):
            r = await coro
            result.append(r)
        return result[0]

    if fyld:
        async for a in \
                asyncyield():
            return await a()

    if pth:
        return await A().a

    if tsk:
        # No se puede crear tarea si es asyncyield
        t = create_task(
            rv_async()
        )
        return await t

    if ens:
        return await asyncio.ensure_future(
            rv_async()
        )

    if gth:
        results = await asyncio.gather(rv_async())
        return results[0]


A.c()

context = asyncio.run(asyncdef(ctx=True), debug=False)
thread = asyncio.run(asyncdef(thrd=True), debug=False)
completed = asyncio.run(asyncdef(cmp=True), debug=False)
foryield = asyncio.run(asyncdef(fyld=True), debug=False)
prop_to_thread = asyncio.run(asyncdef(pth=True), debug=False)
task = asyncio.run(asyncdef(tsk=True), debug=False)
ensure = asyncio.run(asyncdef(ens=True), debug=False)
